//
// glView.cpp
//
// The code maintaining the view for the rendered image
//
#include <math.h>
#include "glView.h"
#include "aniUI.h"
#include "bitmap.h"
#include "arcball.h"
#include "ctrlPanel.h"
#include "Curves.h"

extern ctrlPanel* g_panel;

extern float g_fps;
extern float getCurrentTime();


CGLView::CGLView(int x, int y, int w, int h, const char *l, Scene* scn)
			: Fl_Gl_Window(x,y,w,h,l)
{
	m_nWindowWidth = w;
	m_nWindowHeight = h;

	m_pBuf=NULL;
	m_bSizeChange=true;

	m_pScene=scn;
	user_data((void*)(this));

	callback(cb_exit);
	when(FL_HIDE);

}

void CGLView::cb_exit(Fl_Widget* o, void* v) 
{	
	CGLView* self=(CGLView *)(o->user_data());

	if (self->m_pScene) delete self->m_pScene;

	((CAniUI*)(self->getUI()))->removeView(self);
	self->hide();
}

void CGLView::setScene(Scene *scn)
{
	m_pScene=scn;
}

void CGLView::save_state()
{
	m_pScene->dragging = false;
	m_pScene->current=NONE;
	m_pScene->qdown = m_pScene->qnow;
	m_pScene->vfrom = m_pScene->vto;
	m_pScene->orig_zoom = m_pScene->zoom;
	m_pScene->orig_center = m_pScene->center;
}

int CGLView::handle(int event)
{
	static int downx, downy;
	double *u, *v, len;
	int x=0,y=0;
	switch(event)
	{
	case FL_PUSH:
		if(w() < h())
			m_pScene->radius = w()*.9/2;
		else
			m_pScene->radius = h()*.9/2;
		downx = Fl::event_x();
		downy = h() - Fl::event_y();
		m_pScene->vfrom = MouseOnSphere(downx, downy, w()/2, h()/2, m_pScene->radius);
		m_pScene->vto = m_pScene->vfrom;
		m_pScene->dragging = true;
		m_pScene->current = (State)Fl::event_button();
		if(Fl::event_state(FL_SHIFT))
			m_pScene->current = (State)((int)m_pScene->current + 3);
		break;
	case FL_DRAG:
		if(!m_pScene->dragging)
		{
			m_pScene->dragging = false;
			m_pScene->current = NONE;
			break;
		}

		switch(m_pScene->current)
		{
		case ROTATE:
		case CONSTRAIN_ABOUT_X:
		case CONSTRAIN_ABOUT_Y:
		case CONSTRAIN_ABOUT_Z:
			m_pScene->vto = MouseOnSphere(Fl::event_x(),h() - Fl::event_y(), w()/2, h()/2, m_pScene->radius);
			break;
		case DOLLY:
			x = Fl::event_x();
			y = h() - Fl::event_y();
			Mat4 mat;
			Qt_ToMatrix(Qt_Conj(m_pScene->qnow), mat);
			transpose(mat);
			u = (double  *) mat;
			v = (double  *) &mat[1][0];
			len = sqrt(u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
			u[0]/=len;u[1]/=len;u[2]/=len;
			len = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
			v[0]/=len;v[1]/=len;v[2]/=len;
			m_pScene->center.x+=0.5*u[0]*(m_pScene->zoom*(x-downx))/w();
			m_pScene->center.y+=0.5*u[1]*(m_pScene->zoom*(x-downx))/w();
			m_pScene->center.z+=0.5*u[2]*(m_pScene->zoom*(x-downx))/w();
			m_pScene->center.x+=0.5*v[0]*(m_pScene->zoom*(y-downy))/h();
			m_pScene->center.y+=0.5*v[1]*(m_pScene->zoom*(y-downy))/h();
			m_pScene->center.z+=0.5*v[2]*(m_pScene->zoom*(y-downy))/h();
			downx = x;
			downy = y;
			break;
		case ZOOM:
			y = h() - Fl::event_y();
			m_pScene->zoom = m_pScene->orig_zoom +.1*(y-downy);
			break;
		}
		break;
	case FL_RELEASE:
		if(Fl::event_is_click() || !m_pScene->dragging)
		{
			m_pScene->dragging = false;
			m_pScene->current = NONE;
			break;
		}
		save_state();
		break;
	}
	redraw();

	return 1;
}

void CGLView::draw()
{
	if(!valid())
	{

		m_nWindowWidth=w();
		m_nWindowHeight=h();

		m_bSizeChange=true;

		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();

		gluPerspective( 30.0, (double)m_nWindowWidth/m_nWindowHeight, 1.0, 100.0 );
		glViewport( 0, 0, m_nWindowWidth, m_nWindowHeight );
		glEnable(GL_DEPTH_TEST);

		glClearColor(0.0f, 0.0f, 0.0f, 0.0);

		int drawMode=m_pUI->getDrawMode();
		m_pScene->draw_mode_setup(drawMode);
		glTranslated(0,0,-20);

		glGetDoublev(GL_MODELVIEW_MATRIX, (double*)m_pScene->model);

	}

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// get drawMode and drawQuality from UI
	int drawMode=m_pUI->getDrawMode();
	if (m_pScene)
		m_pScene->redraw(drawMode);

}

void CGLView::refresh()
{
	redraw();
}

void CGLView::resizeWindow(int width, int height)
{
	resize(x(), y(), width, height);
	m_nWindowWidth=w();
	m_nWindowHeight=h();
}

void CGLView::saveImage(char *iname)
{
	if (m_bSizeChange) {
		if (m_pBuf) delete [] m_pBuf;
		m_pBuf=new unsigned char [m_nWindowWidth*m_nWindowHeight*3];
		m_bSizeChange=false;
	}

	make_current();

	glReadBuffer(GL_BACK);

	glPixelStorei( GL_PACK_ALIGNMENT, 1 );
	glPixelStorei( GL_PACK_ROW_LENGTH, m_nWindowWidth );
	
	glReadPixels( 0, 0, m_nWindowWidth, m_nWindowHeight, GL_RGB, GL_UNSIGNED_BYTE, m_pBuf );

	writeBMP(iname, m_nWindowWidth, m_nWindowHeight, m_pBuf); 
}
